#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include "iodefine.h"
#include "ivects.h"
#include "tty.h"
#include "timer.h"
#include "cprintf.h"

#define DESC()	"Port 8.3"
#define INIT()	PORT8.DR.BIT.B3 = 0; PORT8.ICR.BIT.B3 = 1
#define LOW()	PORT8.DDR.BIT.B3 = 1, PORT2.DR.BIT.B0 = 0
#define HIGH()	PORT8.DDR.BIT.B3 = 0, PORT2.DR.BIT.B0 = 1
#define READ()	PORT8.PORT.BIT.B3

/*------------------------------------------------------------------*/

static inline void
wait_ns_1 (int n)
{
  /* This takes 1 cycle for the add, and 3 cycles for the branch.  At
     100 MHz, that's 40 nS per iteration.  */
  __asm__ __volatile__ ("\n\
1:	add	#-1, %0\n\
	bgt	1b"
	 : "=r" (n) : "0" (n));
}

/* Scale each of these to the number of iterations of the above loop,
   rounding up - delays are minimums, not maximums.  */
#define NS 41
#define wait_ns(x) wait_ns_1 ((x + (NS-1)) / NS)
#define wait_us(x) wait_ns_1 (((x*1000) + (NS-1)) / NS)
#define wait_ms(x) wait_ns_1 (((x*1000000) + (NS-1)) / NS)

/*------------------------------------------------------------------*/

static unsigned char crc_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53};

int
crc(unsigned char *data, int len)
{
  int crc = 0;
  while (len--)
    crc = crc_table[crc ^ *data++];
  return crc;
}

/*------------------------------------------------------------------*/

void
ow_w1 ()
{
  LOW ();
  wait_us (5);
  HIGH ();
  wait_us (60);
}

void
ow_w0 ()
{
  LOW ();
  wait_us (50);
  HIGH ();
  wait_us (15);
}

int
ow_read ()
{
  int rv;
  LOW ();
  wait_us (5);
  HIGH ();
  wait_us (5);
  rv = READ ();
  wait_us (55);
  return rv;
}

int
ow_reset ()
{
  int rv;
  LOW ();
  wait_us (500);
  HIGH ();
  wait_us (70);
  rv = READ ();
  wait_us (430);
  return rv;
}

ow_write_byte (int b)
{
  int i;
  for (i=0; i<8; i++)
    if (b & (1 << i))
      ow_w1 ();
    else
      ow_w0 ();
}

int
ow_read_byte ()
{
  int i;
  int rv = 0;
  for (i=0; i<8; i++)
    if (ow_read ())
      rv |= 1 << i;
  return rv;
}

ow_eol()
{
}

void
wait_for_complete()
{
  int i = 3;
  /* Three in case we get a presence pulse.  */
  while (i)
    {
      i--;
      if (ow_read() == 0)
	{
	  /*cput_c('W');*/
	  i = 3;
	}
    }
}

int
s16(unsigned char lsb, unsigned char msb)
{
  int rv;
  rv = lsb + (int)msb * 256;
  return rv;
}

#define RETURN0 {cprintf("\033[31m ret:%d\033[0m\n", __LINE__);return 0;}
/*#define RETURN0 {led_off(GREEN); led_on(RED); return 0;}*/
/*#define RETURN0 return 0;*/

unsigned char onewirebuf[100];
unsigned char *onewireval;
static int rom_is_selected = 0;
unsigned char selected_rom[8];

int
onewire(int len, ...)
{
  int i, rv = 1;
  int do_wait = 0;
  int first=-1;
  va_list a;

  onewireval = onewirebuf;

  if (len < 0)
    {
      len = -len;
      do_wait = 1;
    }

  if (ow_reset())
    RETURN0("no reset");

#if 0
  i = ow_write_byte(0x55);
  if (i != 0x55)
    RETURN0;
  sending_addr = 1;
  for (i=0; i<8; i++)
    {
      if (ow_write_byte(ow_target[i]) != ow_target[i])
	{
	  sending_addr = 0;
	  RETURN0;
	}
    }
  sending_addr = 0;
#else
  i = ow_write_byte(0xCC);
#endif

  va_start (a, len);
  for (i=0; i<len; i++)
    {
      int c = va_arg (a, int);
      if (first == -1)
	first = c;
      if (c == -1)
	{
	  onewireval = onewirebuf + i;
	  break;
	}
      ow_write_byte(c);
    }
  for (; i<len; i++)
    {
      onewirebuf[i] = ow_read_byte();
    }
  if (do_wait)
    {
      /*cput_hex_byte (first);*/
      /*cput_c(' ');*/
      wait_for_complete();
      /*cput_c ('\n');*/
    }
  return rv;
}

#define REDO if (redo == 0) { redo = 1; goto write_page_again; }
#define REDO0 { REDO; RETURN0 }

int
poll_climate ()
{
  int i = 0, p, retry, redo;
  static int page[2][8];

  for (p=0; p<2; p++)
    {
      redo = 0;
    write_page_again:
      /* write page 0, Vad or Vdd */
      if (!onewire(3, 0x4e, 0x00, p&1 ? 0x08 : 0x00))
	RETURN0;

      /* read page 0 */
      if (!onewire (11, 0xbe, 0, -1))
	{
	  /*printf("can't read page 0\n");*/
	  continue;
	}
      if ((onewireval[0] & 0x08) != ((p&1) ? 0x08 : 0x00))
	{
	  /*printf("\033[33mre-writing page 0\033[0m\n");*/
	  REDO0;
	}

      /* save page 0 */
      if (!onewire(-2, 0x48, 0x00))
	REDO0;

      if (p == 0)
	/* initiate T convert */
	if (!onewire(-1, 0x44))
	  REDO0;

      /* initiate V convert */
      if (!onewire(-1, 0xb4))
	REDO0;

      /* recall page 0 */
      if (!onewire(-2, 0xb8, 0))
	REDO0;

#define MAX_RETRIES 4
      for (retry=0; retry < MAX_RETRIES; retry++)
	{
	  /* read page 0 */
	  if (!onewire (11, 0xbe, 0, -1))
	    {
	      /*printf("can't read page 0\n");*/
	      continue;
	    }

	  i = crc(onewireval, 9);
	  if (i != 0)
	    {
	      /*printf("\033[34mbad crc %x %b\033[0m\n", onewireval, 8, i);*/
	      continue;
	    }

	  if ((onewireval[0] & 0x08) != ((p&1) ? 0x08 : 0x00))
	    {
	      /*printf("oops: %x\n", onewireval, 8);*/
	      REDO;
	      RETURN0;
	    }

	  break;
	}
      if (retry == MAX_RETRIES)
	RETURN0;

      for (i=0; i<8; i++)
	page[p][i] = onewireval[i];
    }
  ow_eol();

  int traw = s16(page[0][1], page[0][2]);
  int vad = s16(page[0][3], page[0][4]);
  int vdd = s16(page[1][3], page[1][4]);

  printf("t 0x%04x   vad 0x%04x %d   vdd 0x%04x %d\n", traw, vad, vad, vdd, vdd);

#if 0
  if (vdd > 550 || vdd < 400)
    RETURN0;
  if (vad > vdd - 5)
    RETURN0;
#endif

  if (page[0][2] == 0xff && page[0][1] == 0xff)
    RETURN0;

#if 0
  led_on(GREEN);
  led_off(RED);
#endif

  float temp = page[0][2] + page[0][1] / 256.0;

  temp = temp * 9/5 + 32;

  float humidity = ((((float)vad/(float)vdd) - 0.16) / 0.0062) / (1.093 - 0.0012 * temp);

  float correction = 1.093 - 0.0012 * temp;
  humidity = humidity / correction;


#if 1
  printf("\033[32mTemp 0x%04x %f\tVad 0x%04x %d Vdd 0x%04x %d %f%%\033[0m\n",
	 traw, temp,
	 vad, vad, vdd, vdd, humidity);
#endif
  return 1;
}

/*------------------------------------------------------------------*/

main()
{
  SYSTEM.SCKCR.LONG = 0x00010100;

  PORT2.DDR.BYTE = 0xff;
  PORT2.DR.BYTE = 0;

  tty_init ();
  cprintf("OneWire diagnostic, %s\n", DESC ());

  INIT ();
  HIGH ();

 again:
  tty_puts("press a key to scan bus...\n");
  tty_getc ();

  if (ow_reset ())
    {
      cprintf("no one-wire devices present\n");
      goto again;
    }
  cprintf("one-wire device present\n");

  poll_climate ();

  goto again;
}
