#include "ports.h"
#include "cprintf.h"

/* LEDS are:

   LED0 = P8.3
   LED1 = P8.4
   LED2 = P3.3
   LED3 = P3.4 (P3.6 on RX610 board)

   TxD0	= P0.4
   TxD1	= P2.6
*/

static volatile int timer = 0;

static int8_t junk;

static volatile struct {
  int val;
  int count;
  volatile uint8_t *port;
  uint8_t mask;
  uint8_t unmask;
} led[8] = {
  { 0, 0, DR+8, 0x08, ~0x08 },
  { 0, 0, DR+8, 0x10, ~0x10 },
  { 0, 0, DR+3, 0x08, ~0x08 },
  { 0, 0, DR+3, 0x50, ~0x50 },
  { 0, 0, &junk, 0, 0 },
  { 0, 0, &junk, 0, 0 },
  { 0, 0, &junk, 0, 0 },
  { 0, 0, &junk, 0, 0 }
};

static void __attribute__((interrupt))
tpu0_handler ()
{
  int i;
  timer ++;

  for (i=0; i<8; i++)
    {
      led[i].count += led[i].val;
      if (led[i].count & 0xf0000)
	{
	  led[i].count &= 0xffff;
	  *(led[i].port) &= led[i].unmask;
	}
      else
	{
	  *(led[i].port) |= led[i].mask;
	}
    }
}

static void
tick ()
{
  static int itick = 0;
  while (itick == timer)
    ;
  itick = timer;
}

static int igoals[] = { 0x10, 0x00, 0x01, 0x02, 0x04, 0x08,
			0x10, 0x00, 0x08, 0x04, 0x02, 0x01 };

main()
{
  int goal[8];
  int igoal;
  int i, j;

  /* Power up the peripherals.  */
  MSTPCRA = 0;
  MSTPCRB = 0;
  MSTPCRC = 0;

  DR[8] = 0x00;
  DR[3] = 0x00;

  DDR[3] = 0x58;
  DDR[8] = 0x18;

  /* Timer */
  _vectors[IVEC_TPU0_TGI0A] = tpu0_handler;
  IER0D |= 0x01;
  IPR4C = 5;

  TPU[0].tcr = 0x20; /* TGRA, PCLK = 25 MHz */
  TPU[0].tier = 0x01; /* interrupt on TGRA */
  TPU[0].tgra = 25000; /* should be 1 mS */

  TSTRA |= 0x01; /* start timer */

  set_system_clock_speed ();
  setup_interrupts ();

  tty_init ();
  tty_puts("Throbber!\n");
#if 0
  tty_putc(3);
  tty_putc(3);
  tty_putc(0);
#endif

  igoal = 0;
  for (i=0; i<8; i++)
    goal[i] = 0;

  while (1)
    {
      int again = 0;

      for (i=0; i<8; i++)
	{
	  if (led[i].val < goal[i])
	    led[i].val += 0x101;
	  else if (led[i].val > goal[i])
	    led[i].val -= 0x101;
	  else
	    again ++;
	}
      if (again == 8)
	{
	  igoal ++;
	  j = igoals[igoal % (sizeof(igoals)/sizeof(igoals[0]))];
	  for (i=0; i<8; i++)
	    goal[i] = (j & (1<<i)) ? 0xffff : 0x0000;
#if 0
      cprintf("%w %d : %d %d %d %d : %d %d %d %d\n",
	      timer, igoal,
	      goal[0], goal[1], goal[2], goal[3],
	      led[0].val, led[1].val, led[2].val, led[3].val);
#endif
	}

      tick ();
    }
}
